/*
	File:		CAClip.cpp

	Contains:	Container Application Library source - Clipboard interface

	Written by:	Steve Foley, Greg Ames, David Nelson

	Copyright:	 1994 by Apple Computer, Inc., all rights reserved.

	Change History (most recent first):

		 <6>	 6/20/95	RB		Rewrite
		 <5>	 4/18/95	SJF		Implement Clipboard support
		 <3+>	 2/28/95	SJF		Add debug stuff to API calls
		 <3>	 2/13/95	SJF		Interim checkin to update database
		 <2)	12/15/94	SJF		change somGetGlobalEnvironment to _gpProxyShell->GetGlobalEnvironment
		 <1>	10/30/94	GCA,DHN	All mods to make project compile with no
		 							errors under OpenDoc b1 (with SOM).
		 <0>	10/16/94	SJF		first written

	To Do:
*/

#ifndef _CAUTIL_
#include "CAUtil.h"
#endif

#ifndef _CAERROR_
#include "CAError.h"
#endif

#ifndef _CADOCPRIV_
#include "CADocPriv.h"
#endif

#ifndef _CAFRMUTL_
#include "CAFrmUtl.h"
#endif

#ifndef _CASESSN_
#include "CASessn.h"
#endif

#ifndef SOM_ODClipboard_xh
#include "Clipbd.xh"
#endif

#ifndef SOM_ODStorageUnit_xh
#include <StorageU.xh>
#endif

#ifndef _ODUTILS_
#include <ODUtils.h>
#endif

#ifndef SOM_ODWindowState_xh
#include <WinStat.xh>
#endif

#ifndef SOM_ODPart_xh
#include <Part.xh>
#endif

#ifndef SOM_ODFoci_xh
#include <Foci.xh>
#endif

#ifndef SOM_ODWindow_xh
#include <Window.xh>
#endif

#ifndef _TEMPOBJ_
#include <TempObj.h>
#endif

#pragma segment CALib


//-------------------------------------------------------------------------
// Clipboard
//-------------------------------------------------------------------------

// to paste from the clipboard, 
//	CloneFrameRef(clipbdFrameRef,clipbdDocument,destDocument)
// clipbdFrameRef is cloned from the clipboard to destDocument, and its 
// CAFrameRef within the context of destDocument is returned

// to copy to the clipboard, 
//	CloneFrameRef(sourceFrameRef,sourceDocument,clipbdDocument)
// sourceFrameRef is cloned from sourceDocument to the clipboard, and its
// CAFrameRef within the context of the clipboard is returned

// Use Stream API to write native contents to the ClipDocument

// NOTE this function locks the clipboard for the caller and saves 
//	a copy of the key for later. The actual quantity being returned
//	from this function is an ODDraft*. Documentation says that
//	the storage unit associated with this draft must NOT be disposed
//	of, or released, by the caller.


//-------------------------------------------------------------------------

pascal	CADocumentRef CAGetClipboardDocument( void )
{

	Environment*			ev			= gCASession->GetEV();
	ODArbitrator*			arbitrator	= gCASession->GetArbitrator();;
	ODWindowState*			windowState	= gCASession->GetODSession()->GetWindowState(ev);
	ODClipboard*			clipboard	= kODNULL;
	CAClipboardDocument*	clipDoc		= kODNULL;
	ODTypeToken				clipboardFocus;
	
	CA_TRY
	
		clipboardFocus = gCASession->GetODSession()->Tokenize(ev, kODClipboardFocus);
		TempODWindow activeWindow = windowState->AcquireActiveWindow (ev);
		
	    if (!arbitrator->RequestFocus(ev, clipboardFocus, activeWindow->GetRootFrame(ev)))
			THROW(kCAErrCantGetFocus);

		clipDoc = gCASession->GetCAClipboardDocument();
		
	CA_CATCH_ALL
	CA_ENDTRY
	
	return	((CADocumentRef)clipDoc);
	
}


pascal void CAExportClipboard (void)
{

	Environment*		ev			= gCASession->GetEV();
	ODClipboard*		clipboard	= kODNULL;
	
	clipboard = gCASession->GetODSession()->GetClipboard(ev);
	
	if (clipboard)
	{
		clipboard->PutContentOnPlatformClipboard (ev);
	}

	
}

