/*
	File:		CADataX.cpp

	Contains:	Container Application Library source

	Written by:	Tantek elik & Rick Badertscher

	Copyright:	 1994 by Apple Computer, Inc., all rights reserved.

	Change History (most recent first):

		 <6>	 4/19/95	RB		Moved CARemoveFrameRef() to CALayout.cpp
		 <5>	 4/13/95	RB		Implemented CABeginClone(), CAClone(), CAEndClone()
		 							and CARemoveFrameRef()
		 <4+>	 2/28/95	SJF		Add debug stuff to API calls
		 <4>	 2/13/95	SJF		Interim checkin to update project database
		 <3)	12/15/94	SJF		change somGetGlobalEnvironment to _gpProxyShell->GetGlobalEnvironment
		 <2>	10/30/94	GCA,DHN	All mods to make project compile with no
		 							errors under OpenDoc b1 (with SOM).
		 <1>	  8/9/94	T		first written

	To Do:
*/


#ifndef _CASESSN_
#include "CASessn.h"
#endif

#ifndef _CAUTIL_
#include "CAUtil.h"
#endif

#ifndef _CAERROR_
#include "CAError.h"
#endif

#ifndef _CADEBUG_
#include "CADebug.h"
#endif

#ifndef _CAFRMUTL_
#include "CAFrmUtl.h"
#endif

#ifndef _CADOCPRIV_
#include "CADocPriv.h"
#endif

#ifndef SOM_ODStorageUnit_xh
#include <StorageU.xh>
#endif

#ifndef SOM_ODDraft_xh
#include <Draft.xh>
#endif

#ifndef _ODUTILS_
#include <ODUtils.h>
#endif

#ifndef _STORUTIL_
#include <StorUtil.h>
#endif

#ifndef _StdTypIO_
#include <StdTypIO.h>
#endif

#ifndef SOM_ODPart_xh
#include <Part.xh>
#endif

#ifndef SOM_ODShape_xh
#include <Shape.xh>
#endif

#ifndef _CAPROXYPARTDEF_
#include "CAProxyPartDef.h"
#endif

#ifndef CAProxyExtension_API
#include "CAProxyExtension.xh"
#endif

#ifndef SOM_ODWindow_xh
#include <Window.xh>
#endif

#ifndef _ORDCOLL_
#include <OrdColl.h>
#endif

#ifndef SOM_ODFrame_xh
#include <Frame.xh>
#endif

#ifndef _TEMPOBJ_
#include <TempObj.h>
#endif

#ifndef SOM_Module_OpenDoc_StdTypes_defined
#include <StdTypes.xh>
#endif

#ifndef SOM_Module_OpenDoc_StdProps_defined
#include <StdProps.xh>
#endif

#ifndef SOM_Module_OpenDoc_StdDefs_defined
#include <StdDefs.xh>
#endif

#pragma segment CALib


//-------------------------------------------------------------------------

class CACloneInfo
{

	public:
	
								CACloneInfo(ODDraftKey draftKey);
								~CACloneInfo();
				
		void					AddFrameID (ODID frameID);
		ODULong					GetFrameIDCount (void);
						
		OrderedCollectionIterator*	CreateFrameIterator (void);
		
		Boolean					fCloneComplete;
		ODDraftKey				fDraftKey;

	private:
		
		OrderedCollection*		fFrameIDList;
	
};

//-------------------------------------------------------------------------

static ODFrame*		CreateFrameForPart (		CADocumentRef	document,
												ODID			newPartID,
												ODShape*		frameShape);

static void 		CloneSingleFrameToDocument(	CADocumentRef	fromDocument,	
												CADocumentRef	toDocument,
												CAFrameRef		fromFrameRef,		
												CACloneKey		cloneKey);
												
static CAFrameRef	EmbedPartFromDocument(		CADocumentRef	fromDocument,	
												CADocumentRef	toDocument,
												CACloneKey		cloneKey);
												
static ODID			GetContentFrameID (			CADocumentRef document);


//-------------------------------------------------------------------------
// Data Interchange
//-------------------------------------------------------------------------

pascal	CACloneKey	CABeginClone(	CADocumentRef	sourceDocument,
									CADocumentRef	destDocument,
									CACloneKind		cloneKind )
{
	Environment*	ev			= gCASession->GetEV();
	CACloneInfo*	cloneKey	= kODNULL;
	ODCloneKind		odCloneKind	= kODCloneCopy;
	ODFrame*		destFrame	= kODNULL;
	short			sourceType, destType;
	ODDraft*		sourceDraft;

	ODDraftKey		draftKey;

	sourceType = ((CADocument*) sourceDocument)->GetDocumentType();
	destType = ((CADocument*) destDocument)->GetDocumentType();
	
	if (sourceType == kCAProxyDocument)
		TRACE (3, "CABeginClone - From ProxyDocument");
	else
		TRACE (3, "CABeginClone - From transfer document");

	CA_TRY
	
		// Prepare the CADocuments for cloning (get the content SU from the OD object)
		if (sourceType == kCAClipboardDocument)
			((CAClipboardDocument*)sourceDocument)->PrepareForClone (cloneKind);
		
		if (destType == kCAClipboardDocument)
			((CAClipboardDocument*)destDocument)->PrepareForClone (cloneKind);
		
		if (sourceType == kCADragDropDocument)
			((CADragDropDocument*)sourceDocument)->PrepareForClone (cloneKind);
		
		if (destType == kCADragDropDocument)
			((CADragDropDocument*)destDocument)->PrepareForClone (cloneKind);
		
		sourceDraft = ((CADocument*)sourceDocument)->GetDraft();

		if (destType == kCAProxyDocument)
			destFrame = GetRootFrame (destDocument);

		odCloneKind = (ODCloneKind) cloneKind;
		draftKey = sourceDraft->BeginClone(ev, ((CADocument*)destDocument)->GetDraft(),
												destFrame, odCloneKind);

		cloneKey = new CACloneInfo (draftKey);
		
	CA_CATCH_ALL
	CA_ENDTRY
	
	// Actual ODDraft::BeginClone() is called in CACloneFrameRef

	return (CACloneKey)cloneKey;
}


//-------------------------------------------------------------------------

pascal	void	CAEndClone( CADocumentRef sourceDocument, CADocumentRef destDocument, CACloneKey cloneKey )
{
	Environment*	ev			= gCASession->GetEV();
	CACloneInfo*	cloneInfo	= (CACloneInfo*) cloneKey;	
	ODDraft*		sourceDraft = ((CADocument*)sourceDocument)->GetDraft();
	ODDraft*		destDraft = ((CADocument*)destDocument)->GetDraft();
	short			destType;
	ODFrame*		newFrame;
	
	if (!cloneInfo->fCloneComplete)
		sourceDraft->EndClone (ev, cloneInfo->fDraftKey);

	destType = ((CADocument*) destDocument)->GetDocumentType();

	if (destType == kCAProxyDocument)
	{
	
		OrderedCollectionIterator*	iter;
		ODID						id;
		
		ODFrame* rootFrame = GetRootFrame (destDocument);
		
		if (cloneInfo->GetFrameIDCount() > 0)
		{
		
			iter = cloneInfo->CreateFrameIterator();
			
			for (id = (ODID) iter->First(); iter->IsNotComplete(); id = (ODID) iter->Next())
			{
		

				newFrame = destDraft->AcquireFrame (ev, id);
				
				TRACE (3, "Frame %x new cloned frame", newFrame);

				newFrame->SetContainingFrame (ev, rootFrame);

				if (rootFrame)
				{
					// ChangeLinkStatus fails if the containing frame is NULL
					newFrame->ChangeLinkStatus(ev, kODNotInLink);
				}
				
				newFrame->SetInLimbo (ev, kODFalse);

				// Tell the proxy part about the new frame

				if (((CADocument*) destDocument)->GetDocumentType() == kCAProxyDocument)
					CallProxyAddFrame (destDocument, newFrame);
			
				newFrame->Release(ev);
			}
			
			ODDeleteObject (iter);

		} 		
	
	}
	
	ODDeleteObject (cloneInfo);
	
}

//-------------------------------------------------------------------------

pascal	CAFrameRef CACloneFrameRef( 
	CAFrameRef		fromFrameRef,		
	// can be NULL if only one frame in source document
	CADocumentRef	fromDocument,
	CADocumentRef	toDocument,
	Boolean			isRootFrameRef,
	CACloneKey		cloneKey)
{
	Environment*		ev = gCASession->GetEV();
	ODDraft*			fromDraft	= ((CADocument*)fromDocument)->GetDraft();
	ODDraft*			toDraft		= ((CADocument*)toDocument)->GetDraft();
	CACloneInfo*		cloneInfo	= (CACloneInfo*) cloneKey;	
	ODFrame*			fromFrame	= kODNULL;
	ODFrame*			newFrame	= kODNULL;
	ODStorageUnit*		rootPartSU	= kODNULL;
	ODStorageUnit*		newSU		= kODNULL;
	ODID				newFrameID	= 0;
	ODID				fromFrameID	= 0;
	CAFrameRef			toFrameRef;
	
	// If toDocument isn't a CA document, rootFrame is NULL
	
	if (!fromFrameRef && isRootFrameRef)
	{
	
		ODStorageUnit*		fromPartSU	= kODNULL;
		ODPart*				newPart		= kODNULL;
		ODID				fromPartID	= 0;
		
		if (((CADocument*)toDocument)->GetDocumentType() != kCAProxyDocument)
			DebugStr ("\p CACloneRef error");

		CA_TRY

			toFrameRef = EmbedPartFromDocument (fromDocument, toDocument, cloneKey);
		
		CA_CATCH_ALL
		CA_ENDTRY
		
		if (CAError()) return toFrameRef;

	}
	
	else if (fromFrameRef && isRootFrameRef)
	{
	
		CA_TRY

			CloneSingleFrameToDocument (fromDocument, toDocument, fromFrameRef, cloneKey);
		
		CA_CATCH_ALL
		CA_ENDTRY
		
		fromFrameID = GetIDFromFrameRef (fromDocument, fromFrameRef);

		if (((CADocument*)toDocument)->GetDocumentType() == kCADragDropDocument)
		{
			fromFrame = fromDraft->AcquireFrame (ev, fromFrameID);
			fromFrame->SetDragging (ev, kODTrue);
			//fromFrame->SetInLimbo (ev, kODTrue);
			((CADragDropDocument*)toDocument)->AddDraggingFrame (fromFrame);
			fromFrame->Release(ev);
		}

		toFrameRef = 0;

		if (CAError()) return toFrameRef;

	}
	else
	{
	
		CA_TRY
		
			fromFrameID = GetIDFromFrameRef (fromDocument, fromFrameRef);

			// Force internalization
			fromFrame = fromDraft->AcquireFrame (ev, fromFrameID);
			
 			newFrameID =  fromDraft->Clone (ev,
										cloneInfo->fDraftKey,
										fromFrameID,
										0,
										fromFrameID);

			TRACE (3, "Frame %x being cloned, refCount = %d", fromFrame, fromFrame->GetRefCount(ev));
			
			fromFrame->Release(ev);
			toFrameRef = GetFrameRefFromID (toDocument, newFrameID);
			
			cloneInfo->AddFrameID (newFrameID);
		
		CA_CATCH_ALL
		CA_ENDTRY
		
		if (CAError()) return toFrameRef;
		
		if (((CADocument*)toDocument)->GetDocumentType() == kCADragDropDocument)
		{
			fromFrame = fromDraft->AcquireFrame (ev, fromFrameID);
			fromFrame->SetDragging (ev, kODTrue);
			((CADragDropDocument*)toDocument)->AddDraggingFrame (fromFrame);
			fromFrame->Release(ev);

		}
					

	}
	
	return (toFrameRef);

}

//-------------------------------------------------------------------------

pascal	CAFrameRef	CAGetRootFrameRef (	CADocumentRef	document)
{
	CAFrameRef			rootFrameRef = 0;
	
	CA_TRY
		rootFrameRef = GetRootFrameRef (document);
	CA_CATCH_ALL
	CA_ENDTRY
	
	return rootFrameRef;
	
}

//-------------------------------------------------------------------------

pascal	void	CARemoveFrameRef( CADocumentRef document, CAFrameRef frame )
{
	Environment*		ev			= gCASession->GetEV();
	ODWindowState*		windowState = gCASession->GetODSession()->GetWindowState(ev);
	CAProxyExtension*	proxyExt;
	ODFrame*			theFrame;
	OrderedCollection*	limboFrames;
	
	if (FailIfNotVisible(document)) return;

	CA_TRY
	
		theFrame = AcquireFrameFromFrameRef (document, frame);
		THROW_IF_NULL (theFrame);
		
		TempODPart tempPart = ((CADocument*)document)->AcquireRootPart();

		proxyExt = (CAProxyExtension*)tempPart->AcquireExtension( ev, kCAProxyPartExtension );
		THROW_IF_NULL (proxyExt);
		
		proxyExt->RemoveFrame( ev, theFrame);
		
		limboFrames = ((CADocument*)document)->GetLimboFrames();
		
		// If we set the frame InLimbo, and the frame is still in limbo
		// then remove it, else release it (another part in this draft acquired it)
		if ((limboFrames) && (limboFrames->Contains (theFrame)))
		{
			limboFrames->Remove (theFrame);

			if (theFrame->IsInLimbo (ev))
			{
				TRACE (3, "CARemoveFrameRef - removing cut frame, %x from CADocument, %x", theFrame, document);
				theFrame->Remove(ev);
			}

			else
			{
				TRACE (3, "CARemoveFrameRef - releasing cut frame, %x from CADocument, %x", theFrame, document);
				theFrame->Release(ev);
			}
		}
		else
		{
			TRACE (3, "CARemoveFrameRef - removing frame, %x from CADocument, %x", theFrame, document);
			theFrame->Remove(ev);
		}

	CA_CATCH_ALL
	CA_ENDTRY
	
	ODRelease (ev, proxyExt);
	
	
}

//===========================================================================
#pragma mark '       Private Routines
//===========================================================================


// when copying a bunch of data from one document to another,
// BeginClone, then Clone the FrameRefs, the stream out native content using
// the newly returned FrameRefs, then EndClone
// if none of the Cloned FrameRefs had isRootFrameRef set to true, then EndClone
// sets the native content as the root.


static ODFrame* CreateFrameForPart (CADocumentRef document, ODID newPartID, ODShape* frameShape)
{
	Environment*	ev			= gCASession->GetEV();
	ODShape*		newShape	= kODNULL;
	ODFrame*		rootFrame	= kODNULL;
	ODFrame*		newFrame	= kODNULL;
	ODDraft*		draft		= kODNULL;
	ODPart*			newPart		= kODNULL;


	rootFrame = GetRootFrame (document);

	draft =  ((CADocument*)document)->GetDraft();
	newPart = draft->AcquirePart (ev, newPartID);
	THROW_IF_NULL (newPart);

	// decide what newShape should be - for now, give it what it asked for
	if (frameShape)
	{	
		newShape = frameShape;
		ODAcquireObject (ev, newShape);
	}
	else
	{	// no frameShape specified, use a default size
	
		ODRect rect;
		rect.SetInt(0,0,80,80);
		newShape = rootFrame->CreateShape(ev);
		newShape->SetRectangle(ev, &rect);
	}

	CA_TRY
	
		newFrame = draft->CreateFrame(ev, kODFrameObject, rootFrame,
							newShape, (ODCanvas*)kODNULL, newPart,
							gCASession->GetODSession()->Tokenize(ev, kODViewAsFrame),
							gCASession->GetODSession()->Tokenize(ev, kODPresDefault),
							kODFalse, kODFalse);
	CA_CATCH_ALL
	CA_ENDTRY
	
	ODRelease (ev, newPart);
	ODRelease (ev, newShape);

	return (newFrame);

}


static CAFrameRef		EmbedPartFromDocument(	CADocumentRef	fromDocument,	
												CADocumentRef	toDocument,
												CACloneKey		cloneKey)
{
	Environment*		ev = gCASession->GetEV();
	ODFrame*			newFrame	= kODNULL;
	ODFrame*			rootFrame	= kODNULL;
	ODID				newFrameID	= 0;
	ODDraft*			fromDraft	= ((CADocument*)fromDocument)->GetDraft();
	ODDraft*			toDraft		= ((CADocument*)toDocument)->GetDraft();
	CACloneInfo*		cloneInfo	= (CACloneInfo*) cloneKey;	
	ODID				newPartID;
	ODStorageUnit*		fromPartSU	= kODNULL;
	ODID				fromPartID	= 0;
	ODID				fromFrameID	= 0;
	CAFrameRef			toFrameRef;
	
	// If toDocument isn't a CA document, rootFrame is NULL
	
	rootFrame = GetRootFrame (toDocument);

	// Get the root part id
	
	fromPartSU = ((CADocument*)fromDocument)->GetRootPartSU();
	fromPartID = fromPartSU->GetID(ev);
				
	fromFrameID = GetContentFrameID (fromDocument);

	// Clone the root SU
	newPartID = fromDraft->Clone (ev, cloneInfo->fDraftKey, fromPartID, 0, fromFrameID);

	// Clone the content frame if present

	if (fromFrameID)
	{
		newFrameID = fromDraft->Clone (ev, cloneInfo->fDraftKey, fromFrameID, 0, fromFrameID);
	}
		
		
	fromDraft->EndClone (ev, cloneInfo->fDraftKey);
			
	cloneInfo->fCloneComplete = true;

	// Create a frame for the part if needed
		
	if (!fromFrameID)
	{
		ODShape* frameShape = kODNULL;

		if (fromPartSU->Exists(ev, kODPropFrameShape, kODNULL, 0) )
		{		
			frameShape = rootFrame->CreateShape(ev);
			THROW_IF_NULL (frameShape);
			
			// Focus to the frame shape property
			fromPartSU->Focus(	ev,
								kODPropFrameShape,
								kODPosUndefined, 
								kODNULL, 
								0, 
								kODPosUndefined);
			frameShape->ReadShape(ev, fromPartSU);
		}
			
		newFrame = CreateFrameForPart (toDocument, newPartID, frameShape);
		THROW_IF_NULL (newFrame);
		ODRelease (ev, frameShape);
			
	}
	else 
	{
		newFrame = toDraft->AcquireFrame (ev, newFrameID);
		newFrame->SetInLimbo(ev, kODFalse);

		THROW_IF_NULL (newFrame);
	}
	
	
	newFrame->SetContainingFrame (ev, rootFrame);

	CallProxyAddFrame (toDocument, newFrame);
	
	newFrame->ChangeLinkStatus(ev, kODNotInLink);
	toFrameRef = FrameRefFromFrame (toDocument, newFrame);

	ODRelease (ev, newFrame);
	
	return (toFrameRef);
	
}

static void 	CloneSingleFrameToDocument(		CADocumentRef	fromDocument,	
												CADocumentRef	toDocument,
												CAFrameRef		fromFrameRef,		
												CACloneKey		cloneKey)
{
	Environment*		ev = gCASession->GetEV();
	ODFrame*			fromFrame	= kODNULL;
	ODPart*				fromPart	= kODNULL;
	ODStorageUnit*		toRootSU	= kODNULL;
	ODID				toFrameID	= 0;
	ODID				toRootID	= 0;
	ODDraft*			fromDraft	= ((CADocument*)fromDocument)->GetDraft();
	ODDraft*			toDraft		= ((CADocument*)toDocument)->GetDraft();
	CACloneInfo*		cloneInfo	= (CACloneInfo*) cloneKey;	
	ODID				fromPartID	= 0;
	ODID				fromFrameID	= 0;

	toRootSU = (((CADocument*)toDocument)->GetRootPartSU());
	
	fromFrame = AcquireFrameFromFrameRef (fromDocument, fromFrameRef);
	
	// Set the inLimbo flag for the frame
	if ( (((CADocument*)fromDocument)->GetDocumentType() == kCAProxyDocument) &&
		 (((CADocument*)toDocument)->GetCloneKind() == kCACloneCut))
	{
		fromFrame->SetInLimbo (ev, kODTrue);
		((CADocument*)fromDocument)->AddLimboFrame (fromFrame);
	}
	
	ODSURemoveProperty (ev, toRootSU, kODPropContents);
	toRootSU->AddProperty(ev, kODPropContentFrame);

	// Clone the root SU
	fromPart = fromFrame->AcquirePart(ev);
	toRootID = fromDraft->Clone(	ev,
									cloneInfo->fDraftKey,
									fromPart->GetID(ev),
									toRootSU->GetID(ev),
									fromFrame->GetID(ev));

	fromPart->Release(ev);
	
	// Clone the frame
	toFrameID = fromDraft->Clone(	ev,
									cloneInfo->fDraftKey,
									fromFrame->GetID(ev),
									0,
									fromFrame->GetID(ev));

	fromDraft->EndClone (ev, cloneInfo->fDraftKey);
	cloneInfo->fCloneComplete = true;
	
	// Weakly reference the frame 

	ODSetWeakSURefProp(ev, toRootSU, kODPropContentFrame, kODWeakStorageUnitRef, toFrameID);

	if (((CADocument*)toDocument)->GetDocumentType() == kCADragDropDocument)
	{
		fromFrame->SetDragging (ev, kODTrue);
		((CADragDropDocument*)toDocument)->AddDraggingFrame (fromFrame);
	}

	fromFrame->Release(ev);
	
}


//-------------------------------------------------------------------------

static ODID GetContentFrameID (CADocumentRef document)
{
	Environment*		ev				= gCASession->GetEV();
	ODStorageUnit*		rootPartSU		= kODNULL;
	ODID				contentFrameID	= 0;
	
	rootPartSU = ((CADocument*)document)->GetRootPartSU();
	
	if (rootPartSU->Exists (ev, kODPropContentFrame, kODWeakStorageUnitRef, 0))
	{
		contentFrameID = ODGetWeakSURefProp(ev, rootPartSU, kODPropContentFrame, kODWeakStorageUnitRef);
	}
	else if (rootPartSU->Exists (ev, kODPropDisplayFrames, kODWeakStorageUnitRefs, 0))
	{
		contentFrameID = ODGetWeakSURefProp(ev, rootPartSU, kODPropContentFrame, kODWeakStorageUnitRefs);
	}
	return (contentFrameID);
	
}

//-------------------------------------------------------------------------
CACloneInfo::CACloneInfo(ODDraftKey draftKey)
{

	fFrameIDList		= kODNULL;
	fDraftKey			= draftKey;
	fCloneComplete		= kODFalse;
	
}
						
//-------------------------------------------------------------------------
CACloneInfo::~CACloneInfo()
{

	ODDeleteObject (fFrameIDList);
}

//-------------------------------------------------------------------------
void CACloneInfo::AddFrameID (ODID frameID)
{
	
	if (!fFrameIDList)
	{
		fFrameIDList = new OrderedCollection;
	}
	
	fFrameIDList->AddLast ( (ElementType) frameID);

}

//-------------------------------------------------------------------------
ODULong CACloneInfo::GetFrameIDCount (void)
{
	
	if (!fFrameIDList)
	{
		return 0;
	}
	
	return (fFrameIDList->Count());

}

//-------------------------------------------------------------------------
OrderedCollectionIterator* CACloneInfo::CreateFrameIterator (void)
{
	
	if (!fFrameIDList) return kODNULL;

	return (fFrameIDList->CreateIterator());
	
}

