/*
	File:		CADnD.cpp

	Contains:	Container Application Library source - Drag & Drop interface

	Written by:	Rick Badertscher

	Copyright:	 1994 by Apple Computer, Inc., all rights reserved.

	Change History (most recent first):

		 <13>	 7/14/95	RB		Another rewrite and remove CARemoveDrag* calls (unneccessery)
		 <12>	 6/25/95	RB		Rewrite
		 <11>	  5/8/95	SJF		cleanup from WWDC bug fix session fallout
		 <9>	  5/6/95	SJF		WWDC bug fix session with Rick
		 <7>	 4/18/95	SJF		finish CAGetDragDropDocument()
		 <6>	  4/7/95	SJF		Merge in changes made before the Nice
									Kitchen... (hook up ProxyExtension to do
									some real work)
		 <5>	 3/15/95	SJF		Add in Drag & Drop support
		 <3+>	 2/28/95	SJF		Add debug stuff to API calls
		 <3>	 2/13/95	SJF		Interim checkin to update project database
		 <2)	12/15/94	SJF		change somGetGlobalEnvironment to _gpProxyShell->GetGlobalEnvironment
		 <1>	10/30/94	GCA,DHN	All mods to make project compile with no
		 							errors under OpenDoc b1 (with SOM).
		 <0>	10/16/94	SJF		first written

	To Do:
*/


#ifndef _ALTPOINT_
#include <AltPoint.h>
#endif

#ifndef _CAERROR_
#include "CAError.h"
#endif

#ifndef _CAFRMUTL_
#include "CAFrmUtl.h"
#endif

//#ifndef SOM_ODBaseDragAndDrop_xh
//#include "DragDrpB.xh"
//#endif

#ifndef SOM_ODDragAndDrop_xh
#include "DragDrp.xh"
#endif

//#ifndef SOM_ODBaseDragItemIterator_xh
//#include "DgItmItB.xh"
//#endif

#ifndef SOM_ODDragItemIterator_xh
#include "DgItmIt.xh"
#endif

#ifndef _ODUTILS_
#include <ODUtils.h>
#endif

#ifndef SOM_ODSession_xh
#include <ODSessn.xh>
#endif

#ifndef SOM_ODStorageSystem_xh
#include <ODStor.xh>
#endif

#ifndef SOM_ODContainer_xh
#include <ODCtr.xh>
#endif

#ifndef SOM_ODDocument_xh
#include <Document.xh>
#endif

#ifndef SOM_ODDraft_xh
#include <Draft.xh>
#endif

#ifndef SOM_ODTranslation_xh
#include <Translt.xh>
#endif

#ifndef SOM_Module_OpenDoc_StdDefs_defined
#include <StdDefs.xh>
#endif

#ifndef _DRAGPRIV_
#include "DragPriv.h"
#endif

#ifndef _STORUTIL_
#include <StorUtil.h>
#endif

#ifndef _CAPROXYPARTDEF_
#include "CAProxyPartDef.h"
#endif

#ifndef _CADOCPRIV_
#include "CADocPriv.h"
#endif

#ifndef _CASESSN_
#include "CASessn.h"
#endif

#ifndef CAProxyExtension_API
#include "CAProxyExtension.xh"
#endif

#ifndef _StdTypIO_
#include <StdTypIO.h>
#endif

#ifndef _LINKLIST_
#include <LinkList.h>
#endif

#ifndef SOM_ODPart_xh
#include <Part.xh>
#endif

#ifndef SOM_ODWindow_xh
#include <Window.xh>
#endif

#ifndef SOM_ODWindowState_xh
#include <WinStat.xh>
#endif

#ifndef SOM_ODFacet_xh
#include <Facet.xh>
#endif

#ifndef _PASCLSTR_
#include <PasclStr.h>
#endif

//#ifndef _POUTILS_
//#include <POUtils.h>
//#endif

#ifndef _NMSPCUTL_
#include <NmSpcUtl.h>
#endif

#ifndef _INFOUTIL_
#include <InfoUtil.h>
#endif

#ifndef SOM_ODNameSpaceManager_xh
#include <NmSpcMg.xh>
#endif

#ifndef SOM_ODValueNameSpace_xh
#include <ValueNS.xh>
#endif

#ifndef _TIME
#include <time.h>
#endif

#ifndef _StdTypIO_
#include <StdTypIO.h>
#endif

#ifndef SOM_Module_OpenDoc_StdTypes_defined
#include <StdTypes.xh>
#endif

#ifndef SOM_Module_OpenDoc_StdProps_defined
#include <StdProps.xh>
#endif

#ifndef SOM_Module_OpenDoc_StdDefs_defined
#include <StdDefs.xh>
#endif


#pragma segment CALib

// ISOString equivalents for Apple scrap types
#define		kODAppleTEXT	"Apple:OSType:Scrap:TEXT"
#define		kODApplestyl	"Apple:OSType:Scrap:styl"
#define		kODApplePICT	"Apple:OSType:Scrap:PICT"
#define		kODApplestxt	"Apple:OSType:Scrap:stxt"
#define		kODApplehfs		"Apple:OSType:Scrap:hfs "
#define		kODApplealis	"Apple:OSType:Scrap:alis"
#define		kODApplesect	"Apple:OSType:Scrap:sect"

#define kODBentoFlavor			0x626E746F	// 'bnto'
#define kODPromiseFlavor		0x70726D73	// 'prms'
#define kODPromiseDesc			0x70726D73	// 'prms'
#define kODDragFlavor			0x50415254	// 'PART'
#define kODScrapTypestyl		0x7374796c	// 'styl'
#define kODScrapTypestxt		0x73747874	// 'stxt'
#define kODScrapTypeTEXT		0x54455854	// 'TEXT'
#define kODUserPromiseFlavor	0x66696c65	// 'file'

static void		MakeUniqueFilename(		Environment *ev, ODStorageUnit *su, FSSpec &spec );
static void		CreateNewUntitledFile(	PlatformFile* newFile);
static void		CreateNewFile(			PlatformFile* newFile);
static OSErr	CloneIntoFile(			Environment* ev, ODDragItem* theDragItem, FSSpec dropFSSpec) ;
static OSErr	GetDropFSSpec(			DragReference theDrag, FSSpec &dropFSSpec);
static Boolean	IsTrashFolder(			FSSpec dropFSSpec);
static OSType 	GetStorageUnitOSType(	Environment *ev, ODStorageUnit *su );
static void		SetFinderInfo(			Environment* ev, ODStorageUnit* su, PlatformFile* file);
static void		SetIconFamily(			Environment* ev, ODStorageUnit* fromSU, PlatformFile* file);
static OSErr	CopyToDragItemInChunks(	Environment* ev,
										FlavorType theType,
										ItemReference theItem,
										DragReference theDrag,
										ODDragItem* theDragItem);

pascal OSErr	CALibSendDataProc(
					FlavorType theType,
					void* refCon,
					ItemReference theItem,
					DragReference theDrag);

void PrintDragReference(DragReference dragRef, char* header);
static void InitDateInfo(Environment* ev, ODStorageUnit* su);

DragSendDataUPP		gCASendProc;
DragSendDataUPP		gDragSendProc;
void*				gCASendRefCon;

//-------------------------------------------------------------------------
// Drag and Drop
//-------------------------------------------------------------------------

pascal	CADocumentRef CAGetDragDropDocument(ItemReference theItem)
{

	CADragDropDocument*	dragDoc		= kODNULL;

	CA_TRY
		dragDoc	= gCASession->GetCADragDropDocument();
		dragDoc->Acquire (theItem);
	CA_CATCH_ALL
	CA_ENDTRY
	
	return	(CADocumentRef)dragDoc;
	
}


pascal	OSErr CAInstallTrackingHandler(		DragTrackingHandler trackingHandler,
											WindowPtr			window,
											void*				handlerRefCon )
{
	Environment*		ev			= gCASession->GetEV();
	ODWindowState*		windowState = gCASession->GetODSession()->GetWindowState(ev);
	ODPart*				rootPart	= kODNULL;
	OSErr 				result		= noErr;
	CAProxyExtension*	proxyExt	= kODNULL;
	ODWindow*			odWindow	= kODNULL;
	CADocument*			docRef;
	
	CA_TRY
	
		odWindow = windowState->AcquireODWindow (ev, window);
		if (!odWindow) THROW (kCAErrInvalidParameter);
		
		docRef = gCASession->FindCADocumentRefForWindow (odWindow);
		if (!docRef) THROW (kCAErrCantResolve);

		rootPart = docRef->AcquireRootPart();
		proxyExt = (CAProxyExtension*)rootPart->AcquireExtension( ev, kCAProxyPartExtension );
		result = proxyExt->SetDragTrackingHandler( ev, trackingHandler, handlerRefCon );
		
	CA_CATCH_ALL
	CA_ENDTRY
	
	ODReleaseObject (ev, odWindow);	
	ODReleaseObject (ev, proxyExt);	
	ODReleaseObject (ev, rootPart);	

	return result;
	
}


//-------------------------------------------------------------------------

pascal	OSErr CAInstallReceiveHandler(		DragReceiveHandler	receiveHandler,
											WindowPtr			window,
										  	void*				handlerRefCon )
{
	Environment*		ev			= gCASession->GetEV();
	ODWindowState*		windowState = gCASession->GetODSession()->GetWindowState(ev);
	ODPart*				rootPart	= kODNULL;
	OSErr 				result		= noErr;
	CAProxyExtension*	proxyExt	= kODNULL;
	ODWindow*			odWindow	= kODNULL;
	CADocument*			docRef;
	
	CA_TRY
	
		odWindow = windowState->AcquireODWindow (ev, window);
		if (!odWindow) THROW (kCAErrInvalidParameter);
		
		docRef = gCASession->FindCADocumentRefForWindow (odWindow);
		if (!docRef) THROW (kCAErrCantResolve);
		
		rootPart = docRef->AcquireRootPart();
		proxyExt = (CAProxyExtension*)rootPart->AcquireExtension( ev, kCAProxyPartExtension );
		result = proxyExt->SetDragReceiveHandler( ev, receiveHandler, handlerRefCon );
		docRef->GetWindow()->GetRootFrame(ev)->SetDroppable (ev, true);

	CA_CATCH_ALL
	CA_ENDTRY
	
	ODReleaseObject (ev, odWindow);	
	ODReleaseObject (ev, proxyExt);	
	ODReleaseObject (ev, rootPart);	

	return result;
	
}

//-------------------------------------------------------------------------

pascal OSErr 	CARemoveTrackingHandler( 	DragTrackingHandler trackingHandler,
											WindowPtr			theWindow)
{
	return noErr;
}
										
											
//-------------------------------------------------------------------------

pascal OSErr 	CARemoveReceiveHandler( 	DragReceiveHandler	receiveHandler,
											WindowPtr			theWindow)
{
	return noErr;
}

//-------------------------------------------------------------------------

pascal void CASetDataSendProc(		CADocumentRef	document,
									DragSendDataUPP sendProc,
									void*			refCon)
{

	gCASendProc = sendProc;
	gCASendRefCon = refCon;
	
}

//-------------------------------------------------------------------------
pascal CADropResult CAStartDrag(	CADocumentRef	document,
									EventRecord*	event,
									RgnHandle		dragRgn)
{
	Environment*		ev				= gCASession->GetEV();
	ODFrame*			sourceFrame		= kODNULL;
	CADragDropDocument*	dragDoc;
	ODPart*				destPart		= kODNULL;
	ODDragAndDrop*		dragDrop		= kODNULL;
	ODStorageUnit*		theSU			= kODNULL;
	ODDragLink*			theDragLink		= kODNULL;
	LinkedListIterator*	theIter			= kODNULL;
	ODDragItemIterator*	dragItemIter	= kODNULL;
	OSErr				result			= noErr;
	ODPart*				part			= kODNULL;
	PromiseHFSFlavor    thePromise;
	DragReference		theDrag;
	short				count,index;
	ODType				theISOType;
	ODPlatformType		platformType;
	ODTranslation*		translate;
	
CA_TRY

	dragDoc = gCASession->GetCADragDropDocument();
   	dragDrop = gCASession->GetODSession()->GetDragAndDrop(ev);
	dragDrop->SetDropResult (ev, kODDropFail);
	
	// Begin the drag by creating a new drag reference
	result = NewDrag(&theDrag);
	ASSERTM(result == noErr, result, "Bad result from NewDrag");

	dragDrop->SetDragReference (ev, theDrag);
	
	sourceFrame = GetRootFrame (document);

	// Scan through all items in the drag and add a flavor for each value
	// in the storage unit's kODContents property
	
	translate = gCASession->GetODSession()->GetTranslation(ev);

	// Register a send procedure so we don't have to cache any data with the drag manager
	result = SetDragSendProc(theDrag, gDragSendProc, (void *) dragDrop);

	theIter = new LinkedListIterator(dragDrop->GetDragItemList(ev));
	
	// Create a drag item for the ODDragDrop contentSU;
	
	for (theDragLink = (ODDragLink*)theIter->First(); theDragLink; theDragLink = (ODDragLink*)theIter->Next())
	{
	
		theSU = theDragLink->fItem->fSU;
		
		ODBoolean	userPromise = theSU->Exists(ev, kODPropContents, kODHFSPromise, 0);
		if (userPromise)
		{
			theSU->Focus(ev, kODPropContents, kODPosUndefined, kODHFSPromise, 0, kODPosUndefined);
			ODByteArray ba;
			ODPart* dummySourcePart;
			theSU->GetPromiseValue(ev, kODHFSPromise, 0, sizeof(PromiseHFSFlavor), &ba, &dummySourcePart);
			if (ba._buffer)
			{
				ODBlockMove(ba._buffer, &thePromise, ba._length);
				ODDisposePtr(ba._buffer);
			}
			thePromise.promisedFlavor = kODUserPromiseFlavor;
		}
		else
		{
			// Always promise a file in case somebody wishes to drag this out to the finder
			thePromise.fileType = kODBentoFlavor;
			thePromise.fileCreator = kODShellSignature;
			thePromise.fdFlags = 0;
			thePromise.promisedFlavor = kODBentoFlavor;
		}
		
		// Always promise a file in case somebody wishes to drag this out to the finder

		result = AddDragItemFlavor(theDrag, (ItemReference)(theDragLink->fItem), flavorTypePromiseHFS,
								   &thePromise, sizeof(PromiseHFSFlavor), 0);
		ASSERTM(result == noErr, result, "Bad result from AddDragItemFlavor");

		result = AddDragItemFlavor(theDrag, (ItemReference)(theDragLink->fItem), (userPromise ? kODUserPromiseFlavor : kODBentoFlavor), 0L, 0L, 0);
		ASSERTM(result == noErr, result, "Bad result from AddDragItemFlavor");

		result = SetDragItemBounds(theDrag, (ItemReference)(theDragLink->fItem), &((**dragRgn).rgnBBox));
		ASSERTM(result == noErr, result, "Bad result from SetDragItemBounds");

		theSU->Focus(ev,kODPropContents, kODPosUndefined, 0, 0, kODPosUndefined);
		count = theSU->CountValues(ev);
		for (index = 1; index <= count; ++index)
		{
			theSU->Focus(ev,kODPropContents, kODPosUndefined, 0, index, kODPosUndefined);
			theISOType = theSU->GetType(ev);
			platformType = translate->GetPlatformTypeFromISOType(ev, theISOType);
			ODDisposePtr(theISOType);
			if (!platformType)
				continue;
			result = AddDragItemFlavor(theDrag, (ItemReference)(theDragLink->fItem), platformType, 0L, 0L, 0);
			ASSERTM(result == noErr, result, "Bad result from AddDragItemFlavor");

			// If styled text ('stxt') was just added, also add 'stly' if
			// a 'TEXT' representation will also be written.
			if (     (platformType == kODScrapTypestxt)
				  && theSU->Exists(ev, kODPropContents, kODAppleTEXT, 0)
				  && !theSU->Exists(ev, kODPropContents, kODApplestyl, 0) )
			{
				result = AddDragItemFlavor(theDrag, (ItemReference)(theDragLink->fItem), kODScrapTypestyl, 0L, 0L, 0);
				ASSERTM(result == noErr, result, "Bad result from AddDragItemFlavor");
			}
		}

		// Always add our special flavor to the drag
		result = AddDragItemFlavor(theDrag, (ItemReference)(theDragLink->fItem), kODDragFlavor, 0L, 0L, 0);
		ASSERTM(result == noErr, result, "Bad result from adding kODDragFlavor");

		// Always add our special dummy promise to the drag
		result = AddDragItemFlavor(theDrag, (ItemReference)(theDragLink->fItem), kODPromiseFlavor, 0L, 0L, 0);
		ASSERTM(result == noErr, result, "Bad result from adding kODPromiseFlavor");
	

	}
	
	part = sourceFrame->AcquirePart (ev);
	
	dragDrop->SetSourceFrame (ev, sourceFrame);
	dragDrop->SetSourcePart (ev, part);
	
	dragDrop->SetDestFacet (ev, NULL);
	
	dragItemIter = new ODDragItemIterator();
	dragItemIter->InitDragItemIterator(ev, theIter);

	dragDrop->SetDragItemIterator (ev, dragItemIter);
	dragDoc->SetDragging();
	
	// Do the drag	
	result = TrackDrag(theDrag, event, dragRgn);

	if (dragDrop->GetDestFacet(ev))
	{
		destPart = dragDrop->GetDestFacet(ev)->GetFrame(ev)->AcquirePart(ev);
		destPart->Release(ev); // fornow
	}
	else
		destPart = kODNULL;
	

	DisposeDrag(theDrag);
	
	dragDrop->SetDragReference (ev, kODNULL);
	dragDrop->SetSourceFrame (ev, kODNULL);
	
	part->Release(ev);
	
	dragDrop->SetSourcePart (ev, kODNULL);
	dragDrop->SetDestFacet (ev, kODNULL);
	
CA_CATCH_ALL

	if (theIter)
		delete theIter;
	if (theDrag)
		DisposeDrag(theDrag);
	dragDrop->SetDragReference (ev, kODNULL);

CA_ENDTRY

	ODDropResult dropResult;
	
	dropResult = dragDrop->GetDropResult(ev);
	
	if (dragDrop->GetDropResult(ev) == kODDropMove)
		return (CADropResult) kODDropMove;
	else if (dragDrop->GetDropResult(ev) == kODDropCopy)
		return (CADropResult)kODDropCopy;
	else
		return (CADropResult)kODDropFail;

	
}

//===========================================================================
#pragma mark '       Private Routines
//===========================================================================

pascal OSErr  CALibSendDataProc(	FlavorType		theType,
								void*			refCon,
								ItemReference	theItem,
								DragReference	theDrag)
{
	Environment*		ev			= gCASession->GetEV();
	ODDragAndDrop*		dragDrop	= (ODDragAndDrop*)refCon;
	OSErr				result		= noErr;
	ODType				theISOType;
	ODTranslation*		translate;
	ODPtr				stylPtr		= kODNULL;
	ODULong				stylSize	= 0;
	ODULong				dropResult;

CA_TRY

	LinkedListIterator iter(dragDrop->GetDragItemList(ev));
	ODDragLink*			theDragLink = kODNULL;
	ODMemDragItem*		theDragItem = kODNULL;

	for (theDragLink = (ODDragLink*)iter.First(); 
	     theDragLink; theDragLink = (ODDragLink*)iter.Next())
	{
		if (theDragLink->fItem == (ODDragItem*)theItem)
		{
			theDragItem = (ODMemDragItem*) theDragLink->fItem;
			break;
		}
	}
	
	
	if (!theDragItem)
	{
		WARN("Item not found in send proc");
		result =  badDragFlavorErr;
	}
	
	else
	{
	
		switch (theType)
		{
			case kODDragFlavor:
				WARN("Send procedure called for type 'PART'");
				result = badDragFlavorErr;
			break;	
			
			case kODPromiseFlavor:
				WARN("Send procedure called for type kODPromiseFlavor");
				result = badDragFlavorErr;
			break;
			
			case kODBentoFlavor:
			{

				// Check the drop location (probably from the finder)
				FSSpec dropFSSpec;
				
				result = GetDropFSSpec(theDrag, dropFSSpec);
				MakeUniqueFilename(ev,theDragItem->fSU, dropFSSpec);
				
				// Drags to the Trash are always a move

				if( IsTrashFolder(dropFSSpec) )
					dropResult = kODDropMove;		// Dragged to Trash, so it's a move
				else
					dropResult = kODDropCopy;
				
				// First give the CA a chance to fulfill the promise
				
				if (gCASendProc)
				{
					result = CallDragSendDataProc(gCASendProc, theType, gCASendRefCon, theDragItem, theDrag);
					if (result == noErr)
					{
						dragDrop->SetDropResult(ev, dropResult);
						break;
					}
				}

				// Create the new file and add the container data.
				
				if ((result = CloneIntoFile(ev, theDragItem, dropFSSpec)) == noErr)
				{
					result = SetDragItemFlavorData(theDrag, theItem, theType, &dropFSSpec, sizeof(FSSpec), 0L);
				}
				else
				{
					dropResult = kODDropFail;
				}				
				dragDrop->SetDropResult(ev,dropResult);
				
			}
			break;
			
			case kODUserPromiseFlavor:
 
 				result = noErr;
				ODStorageUnit* contentSU = theDragItem->fSU;
				if ((contentSU != kODNULL) && (contentSU->Exists(ev, kODPropContents, kODHFSPromise, 0)))
				{
				
					if (result == noErr)
					{
						if (gCASendProc)
						{
							result = CallDragSendDataProc(gCASendProc, theType, gCASendRefCon, theItem, theDrag);
						}
						else
						{
							WARN("No DragSendProc installed.");
							result = badDragFlavorErr;
						}
					}
					
					if (result != noErr) dropResult = kODDropFail;
					dragDrop->SetDropResult(ev, dropResult);

				}
			break;

			default:
			
				// Otherwise fulfill the data by copying from the drag storage unit
				dragDrop->SetDropResult(ev, kODDropFail);
				translate = dragDrop->GetSession(ev)->GetTranslation(ev);
				theISOType = translate->GetISOTypeFromPlatformType(ev, theType, kODPlatformDataType);
				if (theDragItem->fSU->Exists(ev,kODPropContents, theISOType, 0))
				{
					theDragItem->fSU->Focus(ev,kODPropContents, kODPosUndefined, theISOType, 
					                      0, kODPosUndefined);
					
					
					//PrintDragReference(theDrag, "About to CopyToDrag");
					result = CopyToDragItemInChunks(ev, theType, theItem, theDrag, theDragItem);
					if (result == noErr)
						dragDrop->SetDropResult(ev, kODDropCopy);
				}
				else if ( (theType == kODScrapTypestyl) && StorageUnitGetStylFromStyledText(theDragItem->fSU, ev, &stylSize, &stylPtr) )
				{
					result = SetDragItemFlavorData(theDrag, theItem, theType, stylPtr, stylSize, 0L);
					WASSERTM(result == noErr, "Error from SetDragItemFlavorData in send procedure");		
					if (result == noErr)
						dragDrop->SetDropResult(ev, kODDropCopy);
					ODDisposePtr(stylPtr);
					stylPtr = kODNULL;
				}
				else
				{
					WARN("Send procedure can not supply this type");		
					result = badDragFlavorErr;
				}
				ODDisposePtr(theISOType);
					
			break;
			
		}
		
	}

CA_CATCH_ALL
CA_ENDTRY

	return result;
	
}




static void MakeUniqueFilename( Environment *ev, ODStorageUnit *su, FSSpec &spec )
{
	Str255 pname;
	ODIText* name = ODGetITextProp(ev, su, kODPropName, kODMacIText, kODNULL);
	if (name)
		GetITextPString(name, pname);
	else
		pname[0] = 0;

	DisposeIText(name);
	ODBlockMove(pname, spec.name, pname[0]+1);
	
	PlatformFile*	file = new PlatformFile;
	file->Specify(&spec);
	
	if ((pname[0] == 0) || (file->Exists())) {
		CreateNewUntitledFile(file);
	}

	spec = file->GetFileSpec();
	delete file;
}

static void	CreateNewUntitledFile(PlatformFile* newFile)
{
	ODULong	secs;
	GetDateTime(&secs);
	
	ODSByte	dateString[kODMaxFileNameSize];
	IUDateString(secs,shortDate,(StringPtr)dateString);
	PascalToCString((StringPtr)dateString);
	
	ODSByte	untitledName[kODMaxFileNameSize];
	strcpy(untitledName,"untitled ");
	strcat(untitledName,dateString);
	newFile->SetAsciiName(untitledName);
	
	CreateNewFile(newFile);
}

static void	CreateNewFile(PlatformFile* newFile)
{
	const ODULong kMaxNumberSuffixLength = 10;
	ODSLong	seed = 0;
	ODUShort len = newFile->GetAsciiNameLength();

	ODSByte	 	prefixName[kODMaxFileNameSize];
	newFile->GetAsciiName(prefixName,kODMaxFileNameSize);

	ODSByte	 	newName[kODMaxFileNameSize];
	strcpy(newName,prefixName);
	
	do {
		if (seed<0) // overflow, remove one character from prefix
		{
			if (--len == 0)
				THROW(kODErrUndefined);
			seed = 0;
			prefixName[len] = '\0';
			strcpy(newName,prefixName);
		}
		if (seed++>0)
		{
			ODSByte theNum[kMaxNumberSuffixLength];
			
			NumToString(seed, (StringPtr)theNum); // Macintosh Specific
			PascalToCString((StringPtr)theNum);
			
			strcpy(newName,prefixName);
			strncat(newName," ",kODMaxFileNameSize-strlen(newName));
			strncat(newName,theNum,kODMaxFileNameSize-strlen(newName));
		}
		newFile->SetAsciiName(newName);
	} while (newFile->Exists());
}	// CreateNewFile


static OSErr CloneIntoFile(Environment* ev, ODDragItem* theDragItem, FSSpec dropFSSpec) 
{
	OSErr			err = noErr;
	PlatformFile	file;
	file.Specify(&dropFSSpec);
	
	ODContainer* targetContainer = kODNULL;
	ODDocument* targetDocument = kODNULL;
	ODDraft* targetDraft = kODNULL;
	ODStorageUnit* targetDraftProperties	= kODNULL;
	ODStorageUnit* targetSU	 = kODNULL;
	
	ODVolatile(targetContainer);		
	ODVolatile(targetDocument);		
	ODVolatile(targetDraft);		
	ODVolatile(targetDraftProperties);		
	ODVolatile(targetSU);		
	CA_TRY
		ODStorageUnit* contentSU = theDragItem->fSU;
		ODID targetContentFrameID = kODNULLID;
		OSType	fileType = GetStorageUnitOSType(ev,contentSU);

		file.Create(kODShellSignature, fileType, smSystemScript);
		SetFinderInfo(ev, contentSU, &file);
		SetIconFamily(ev, contentSU, &file);
		targetContainer = CreateFileContainer(ev, contentSU->GetSession(ev), &file.GetFileSpec());
		targetDocument = targetContainer->AcquireDocument(ev, kODDefaultDocument);
		targetDraft = targetDocument->AcquireBaseDraft(ev, kODDPExclusiveWrite);
		targetDraftProperties = targetDraft->AcquireDraftProperties(ev);
		targetSU = targetDraft->CreateStorageUnit(ev);
		targetDraftProperties->AddProperty(ev, kODPropRootPartSU)->AddValue(ev,kODStrongStorageUnitRef);
		ODStorageUnitRef ref;
		targetDraftProperties->GetStrongStorageUnitRef(ev, targetSU->GetID(ev), ref);
		StorageUnitSetValue(targetDraftProperties, ev, sizeof(ODStorageUnitRef), ref);
		ODDraft* sourceDraft = contentSU->GetDraft(ev);				
		ODDraftKey cloneKey = 0;
		CA_TRY
			cloneKey = sourceDraft->BeginClone(ev, targetDraft, kODNULL, kODCloneDropCopy);
			sourceDraft->Clone(ev, cloneKey, contentSU->GetID(ev), targetSU->GetID(ev), 0);				
			if (contentSU->Exists(ev, kODPropContentFrame, kODWeakStorageUnitRef, 0)) {
				contentSU->Focus(ev, kODPropContentFrame, kODPosUndefined, kODWeakStorageUnitRef, 0, kODPosUndefined);
				ODID contentFrameID = ODGetWeakSURefProp(ev, contentSU, kODPropContentFrame, kODWeakStorageUnitRef);
				targetContentFrameID = sourceDraft->Clone(ev, cloneKey, contentFrameID, kODNULLID, 0);				
			}		
			sourceDraft->EndClone(ev, cloneKey);
		CA_CATCH_ALL
			sourceDraft->AbortClone(ev, cloneKey);
		CA_ENDTRY
		
		if (targetContentFrameID != kODNULLID) {
			targetDraftProperties->AddProperty(ev,kODPropRootFrameList);
			targetDraftProperties->AddValue(ev,kODStrongStorageUnitRefs);
			ODStorageUnitRef suRef;
			targetDraftProperties->GetStrongStorageUnitRef(ev, targetContentFrameID, suRef);
			StorageUnitSetValue(targetDraftProperties, ev, sizeof(ODStorageUnitRef), suRef);
		}

		if (targetSU->Exists(ev, kODPropMouseDownOffset, kODNULL, 0) == kODTrue) {
			targetSU->Focus(ev, kODPropMouseDownOffset, kODPosUndefined, kODNULL, 0, kODPosAll);
			targetSU->Remove(ev);
		}
		ODReleaseObject(ev, targetDraftProperties);
		ODReleaseObject(ev, targetSU);
		targetDraft->Externalize(ev);
		ODReleaseObject(ev, targetDraft);
		ODReleaseObject(ev, targetDocument);
		ODReleaseObject(ev, targetContainer);

				
	CA_CATCH_ALL
	
		ODReleaseObject(ev, targetDraftProperties);
		ODReleaseObject(ev, targetSU);
		ODReleaseObject(ev, targetDraft);
		ODReleaseObject(ev, targetDocument);
		ODReleaseObject(ev, targetContainer);
		file.Delete();		
		err = dragNotAcceptedErr;
		WARN("Error caught in CloneIntoFile.");
	CA_ENDTRY
	
	return err;
}

static OSErr GetDropFSSpec(ODPlatformDragReference theDrag, FSSpec &dropFSSpec)
{
	AEDesc	dropDescriptor;

	OSErr result = GetDropLocation(theDrag, &dropDescriptor);
	if (result != noErr || dropDescriptor.descriptorType == typeNull)
		result = fnfErr;			// Can't create a file unless we know where it goes!
	else
	{
		AEDesc fssDescriptor;
		result = AECoerceDesc(&dropDescriptor, typeFSS, &fssDescriptor);
		AEDisposeDesc(&dropDescriptor);
		if (result == noErr) {		
			ODBlockMove(*fssDescriptor.dataHandle, &dropFSSpec, sizeof(FSSpec));
			AEDisposeDesc(&fssDescriptor);

			// Use PBGetCatInfo to get the directoryID of the target directory.
			CInfoPBRec getInfoPB;
			getInfoPB.dirInfo.ioCompletion = 0L;
			getInfoPB.dirInfo.ioNamePtr = (StringPtr) &(dropFSSpec.name);
			getInfoPB.dirInfo.ioVRefNum = dropFSSpec.vRefNum;
			getInfoPB.dirInfo.ioFDirIndex = 0;
			getInfoPB.dirInfo.ioDrDirID = dropFSSpec.parID;
			result = PBGetCatInfoSync(&getInfoPB);
			if (result == noErr)
			{
				dropFSSpec.parID = getInfoPB.dirInfo.ioDrDirID;
				dropFSSpec.name[0] = 0;
			}
		}
	}
	
	return result;
}


static Boolean IsTrashFolder(FSSpec dropFSSpec)
{
	short trashVol;
	long trashDir;

	return (FindFolder(dropFSSpec.vRefNum,kTrashFolderType,kDontCreateFolder,&trashVol,&trashDir) == noErr) && 
			(trashVol==dropFSSpec.vRefNum) && (
			(trashDir==dropFSSpec.parID) );

}

static void	SetIconFamily(Environment* ev, ODStorageUnit* fromSU, PlatformFile* file)
{
	static const ResType			kIconType[6] = {'ICN#','icl4','icl8','ics#','ics4','ics8'};
	static const short				kIconSize[6] = {256,   512,   1024,  64,    128,   256   };
 	ODIconFamily icons = ODGetIconFamilyProp(ev, fromSU, kODPropCustomIcon, kODIconFamily, kAllIconsMask);
	
	if (icons)
	{
		// write icons to the file
		long	i = 6;
		do {
			Handle	icon;
			char	state;
			if ((GetIconFromSuite(&icon, icons, kIconType[--i]) == noErr) && (icon != nil))
			{
				state = HGetState(icon);
				HLock(icon);
				WASSERT(GetHandleSize(icon) == kIconSize[i]);
				file->WriteResourcePtr(kIconType[i], kCustomIconResource, *icon, kIconSize[i]);
				HSetState(icon, state);
			}
		} while (i);
		file->SetCustomIcon(kODTrue);
	}
}

static void SetFinderInfo(Environment* ev, ODStorageUnit* su, PlatformFile* file)
{
	InitDateInfo(ev, su);
	ODULong modDate = ODGetModificationDate(ev, su);
	ODULong creationDate = ODGetCreationDate(ev, su);		
	
	file->CreateResFile();	
	file->SetFileModDate(modDate);
	file->SetFileCreationDate(creationDate);
	file->SetStationery(ODGetSUIsStationery(ev, su));
}

static OSType GetStorageUnitOSType( Environment *ev, ODStorageUnit *su )
{
	OSType	returnType = kODShellSignature;

#if ODDebug_DragAndDrop
	ODULong	numProperties = su->CountProperties(ev);
	su->Focus(ev, kODNULL, kODPosAll, kODNULL, 0, kODPosAll);
	for (ODULong i = 1; i <= numProperties; i++) {
		su->Focus(ev, kODNULL, kODPosNextSib,kODNULL,0,kODPosAll);
		ODPropertyName name = su->GetProperty(ev);
		ODDisposePtr(name);
	}
#endif
	
	if (su->Exists(ev, kODPropContents,kODNULL,1)) {
		su->Focus(ev, kODPropContents,kODPosUndefined,kODNULL,1,kODPosUndefined);
		ODType suType = su->GetType(ev);
		if(suType != kODNULL) {
			ODValueNameSpace* osTypeNameSpace = (ODValueNameSpace*)
						su->GetSession(ev)->GetNameSpaceManager(ev)->HasNameSpace(ev,kODKindOldMacOSType);
			if (osTypeNameSpace) {
				ODOSType*   osTypePtr;
				ODOSType	osType;
				ODULong		valueLen;
				ODBoolean	osTypeFound = ValueNameSpaceGetEntry(
											osTypeNameSpace, 
									   ev, 
									   suType, 
										  	(ODPtr*) &osTypePtr,
										    &valueLen);
				if (osTypeFound) {
					osType = *osTypePtr;
					ODDeleteObject( osTypePtr );
					
					WASSERT(valueLen==sizeof(OSType));
					returnType = osType;
				}
			}
		}
	}
	return returnType;
}

static OSErr CopyToDragItemInChunks(Environment* ev,
								FlavorType theType,
								ItemReference theItem,
								DragReference theDrag,
								ODDragItem* theDragItem) 
{
	const ODULong kCopyBlockSize = 512;
	OSErr	result = noErr;
	ODPtr 	dataPtr = ODNewPtr(kCopyBlockSize);
	ODULong dataSize = theDragItem->fSU->GetSize(ev);
	
	ODULong currentByte = 0;
	while ((currentByte < dataSize) && (result == noErr))
	{
		ODULong bytesRead = StorageUnitGetValue(theDragItem->fSU, ev, kCopyBlockSize, dataPtr);
		result = SetDragItemFlavorData(theDrag, theItem, theType, dataPtr, bytesRead, currentByte);
		WASSERTM(result == noErr, "Error from SetDragItemFlavorData in CopyToDragItemInChunks.");		
		currentByte += bytesRead;
	}
	ODDisposePtr(dataPtr);
	WASSERTM(currentByte == dataSize, "Bytes written is NOT equal to dataSize.");		
	
	return result;
}

void PrintDragReference(DragReference dragRef, char* header)
{
	unsigned short	items = 0;
	CountDragItems(dragRef, &items);
	somPrintf("%s: fDragReference %x numItems %d\n", header, dragRef, items);
	for (unsigned short i = 1; i <= items; i++) {
		ItemReference	itemRef;
		GetDragItemReferenceNumber(dragRef, i, &itemRef);
		unsigned short numFlavors = 0;
		CountDragItemFlavors(dragRef, itemRef, &numFlavors);
		FlavorType theType;
		Size	dataSize = 0;
		for (unsigned short j = 1; j <= numFlavors; j++) {
			GetFlavorType (dragRef, itemRef, j, &theType);
			GetFlavorDataSize(dragRef, itemRef, theType, &dataSize);
			somPrintf("%s: itemRef %x item %d flavor %d size %d type ", header, itemRef, i, j, dataSize);
			for (short k = 0; k <= 3; k++) {
				somPrintf("%c", ((char*) &theType)[k]);
			}
			somPrintf("\n");
			if ((theType == 'hfs ') && (dataSize > 0)) {
				HFSFlavor hfsFlavor;
				GetFlavorData(dragRef,itemRef,theType,&hfsFlavor,&dataSize,0);
				somPrintf("type %x creator %x name %s\n", hfsFlavor.fileType, hfsFlavor.fileCreator, p2cstr(hfsFlavor.fileSpec.name));
			}
		}
	}
}

// Copied from POUtils.cpp
static void InitDateInfo(Environment* ev, ODStorageUnit* su)
{
	
	if (su->Exists(ev, kODPropCreateDate, kODTime_T, 0) == kODFalse) 
	{
		ODULong creationTimeDate;
		time((time_t *)(&creationTimeDate));
		ODSetTime_TProp(ev, su, kODPropCreateDate, kODTime_T, creationTimeDate);
		ODSetTime_TProp(ev, su, kODPropModDate, kODTime_T, creationTimeDate);
	}
}

