/*
	File:		CADoc.cpp

	Contains:	Container Application Library source

	Written by: Rick Badertscher, Tantek elik, David Nelson

	Copyright:	 1994 by Apple Computer, Inc., all rights reserved.

	Change History (most recent first):

		 <5>	 6/18/95	RB		Incorporated CADocument usage
		 <5>	 5/26/95	RB		Modified all routines to call GetFocusedStorageUnit().
		 							Added GetFocusedStorageUnit().
		 <4>	 3/16/95	RB		Modified routines to call GetDocumentStreamSUFromDocumentRef()
		 <3+>	 2/28/95	SJF		Add debug stuff to API calls
		 <3>	 2/13/95	SJF		Interim checkin to update project database
		 <2)	12/15/94	SJF		change somGetGlobalEnvironment to _gpProxyShell->GetGlobalEnvironment
		 <1>	10/30/94	GCA,DHN	All mods to make project compile with no
		 							errors under OpenDoc b1 (with SOM).
		 <0>	  8/9/94	T		first written
		 
	To Do:
*/

#ifndef _CASESSN_
#include "CASessn.h"
#endif

#ifndef _CAERROR_
#include "CAError.h"
#endif

#ifndef _CADOCPRIV_
#include "CADocPriv.h"
#endif

#ifndef SOM_ODStorageUnit_xh
#include <StorageU.xh>
#endif

#ifndef _ODUTILS_
#include <ODUtils.h>
#endif

#ifndef _STORUTIL_
#include <StorUtil.h>
#endif

#ifndef _TEMPOBJ_
#include <TempObj.h>
#endif

#pragma segment CALib

//-------------------------------------------------------------------------
// Stream
//-------------------------------------------------------------------------


pascal	void	CARead(CADocumentRef document, char* buffer,CASize* length )
{
	Environment*	ev = gCASession->GetEV();
	ODStorageUnit*	rootSU;

	rootSU = ((CADocument*)document)->GetRootPartSU();

	CA_TRY
	
		if (!rootSU) THROW (kCAErrCantFocusDocument);

		((CADocument*)document)->FocusContentStream();
		
		StorageUnitGetValue(rootSU, ev, *length, (ODValue)buffer);
		((CADocument*)document)->SetStreamOffset (rootSU->GetOffset(ev));
		
	CA_CATCH_ALL
	CA_ENDTRY
	
}

//-------------------------------------------------------------------------

pascal	void	CAWrite(CADocumentRef document,char* buffer,CASize* length )
{
	Environment*	ev = gCASession->GetEV();
	ODStorageUnit*	rootSU;
	ODByteArray		ba;

	rootSU = ((CADocument*)document)->GetRootPartSU();
	
	CA_TRY
	
		if (!rootSU) THROW (kCAErrCantFocusDocument);

		((CADocument*)document)->FocusContentStream();
	
		ba._buffer = (octet*) buffer;
		ba._length = *length;
		ba._maximum = *length;

		rootSU->SetValue (ev, &ba);
		((CADocument*)document)->SetStreamOffset (rootSU->GetOffset(ev));
	
	CA_CATCH_ALL
	CA_ENDTRY
	
}


//-------------------------------------------------------------------------

pascal	CASize	CAGetSize(CADocumentRef document)
{
	Environment*	ev = gCASession->GetEV();
	CASize			size = 0;
	ODStorageUnit*	rootSU;
	
	rootSU = ((CADocument*)document)->GetRootPartSU();
	
	CA_TRY
	
		if (!rootSU) THROW (kCAErrCantFocusDocument);

		((CADocument*)document)->FocusContentStream();
		rootSU = ((CADocument*)document)->GetRootPartSU();
		size = rootSU->GetSize(ev);
	
	CA_CATCH_ALL
	CA_ENDTRY
	
	return (size);
	
}

//-------------------------------------------------------------------------

pascal	CASize	CAGetOffset(CADocumentRef document)
{
	CASize			offset = 0;

	CA_TRY
		offset = ((CADocument*)document)->GetStreamOffset();
	CA_CATCH_ALL
	CA_ENDTRY

	return (offset);
	
}

//-------------------------------------------------------------------------

pascal	void	CASetOffset(CADocumentRef document, CASize offset )
{
	Environment*	ev = gCASession->GetEV();
	ODStorageUnit*	rootSU;

	rootSU = ((CADocument*)document)->GetRootPartSU();

	CA_TRY

		if (!rootSU) THROW (kCAErrCantFocusDocument);

		((CADocument*)document)->FocusContentStream();

		rootSU->SetOffset (ev, offset);
		((CADocument*)document)->SetStreamOffset (offset);
	
	CA_CATCH_ALL
	CA_ENDTRY

}

//-------------------------------------------------------------------------

pascal	void	CAInsert(CADocumentRef document,char* buffer,CASize* length )
{
	Environment*	ev = gCASession->GetEV();
	ODStorageUnit*	rootSU;
	ODByteArray		ba;

	rootSU = ((CADocument*)document)->GetRootPartSU();

	CA_TRY

		if (!rootSU) THROW (kCAErrCantFocusDocument);

		((CADocument*)document)->FocusContentStream();
		rootSU = ((CADocument*)document)->GetRootPartSU();

		ba._length = *length;
		ba._maximum = *length;
		ba._buffer = (octet*) buffer;

		rootSU->InsertValue (ev, &ba);
		((CADocument*)document)->SetStreamOffset (rootSU->GetOffset(ev));
	
	CA_CATCH_ALL
	CA_ENDTRY
	
}

//-------------------------------------------------------------------------

pascal	void	CADelete(CADocumentRef document,char* buffer,CASize* length )
{
	Environment*		ev = gCASession->GetEV();
	ODStorageUnit*		rootSU;
	
	rootSU = ((CADocument*)document)->GetRootPartSU();

	CA_TRY

		if (!rootSU) THROW (kCAErrCantFocusDocument);

		((CADocument*)document)->FocusContentStream();
	
		rootSU = ((CADocument*)document)->GetRootPartSU();
		rootSU->DeleteValue (ev, *length);
		((CADocument*)document)->SetStreamOffset (rootSU->GetOffset(ev));

	CA_CATCH_ALL
	CA_ENDTRY
	
}


